-- 4 tap FIR filter TMC2246
-- dd 12/5/95 
library ieee;
use ieee.std_logic_1164.all;
library exemplar;
use exemplar.exemplar_1164.all;

entity fir is
    port (
		d1 : in std_logic_vector(9 downto 0);
		d2 : in std_logic_vector(9 downto 0);
		d3 : in std_logic_vector(9 downto 0);
		d4 : in std_logic_vector(9 downto 0);
		c1 : in std_logic_vector(10 downto 0);
		c2 : in std_logic_vector(10 downto 0);
		c3 : in std_logic_vector(10 downto 0);
		c4 : in std_logic_vector(10 downto 0);
		clk,fsel,ensel,enb1,enb2,enb3,enb4,acc,not_ocen,not_oen : in std_logic;
		s : out std_logic_vector(15 downto 0)	
   );
end fir;

library ieee;
use ieee.std_logic_1164.all;
library exemplar;
use exemplar.exemplar_1164.all;


architecture operate_fir of fir is

signal		ld1 : std_logic_vector(9 downto 0);
signal		ld2 : std_logic_vector(9 downto 0);
signal		ld3 : std_logic_vector(9 downto 0);
signal		ld4 : std_logic_vector(9 downto 0);
signal		lc1 : std_logic_vector(10 downto 0);
signal		lc2 : std_logic_vector(10 downto 0);
signal		lc3 : std_logic_vector(10 downto 0);
signal		lc4 : std_logic_vector(10 downto 0);
signal		ldtc1,ldtc2,ldtc3,ldtc4 : integer;
signal isum,isum2 : integer;
signal isum3 : integer;
signal temp : integer;
signal		rld1,rld2,rld3,rld4,rlc1,rlc2,rlc3,rlc4,rldtc1,rldtc2,rldtc3,rldtc4 : real;
signal risum,risum2 : real;
SIGNAL  lacc,lacc2,lfsel,lfsel2,lfsel3,lfsel4 : Std_logic;
SIGNAL  clkd1,clkd2,clkd3,clkd4,clkc1,clkc2,clkc3,clkc4 : Std_logic;

constant tD : time := 15 ns;
constant tS : time := 10 ns;
constant tH : time := 2 ns;
constant tDIS : time := 20 ns;



begin

    clkd1 <= (not enb1 and clk) or (enb1 and not ensel and clk);
    clkd2 <= (not enb2 and clk) or (enb2 and not ensel and clk);
    clkd3 <= (not enb3 and clk) or (enb3 and not ensel and clk);
    clkd4 <= (not enb4 and clk) or (enb4 and not ensel and clk);

    clkc1 <= (not enb1 and clk) or (enb1 and ensel and clk);
    clkc2 <= (not enb2 and clk) or (enb2 and ensel and clk);
    clkc3 <= (not enb3 and clk) or (enb3 and ensel and clk);
    clkc4 <= (not enb4 and clk) or (enb4 and ensel and clk);

    dff_v(data=>d1,clk=>clkd1,q=>ld1);
    dff_v(data=>d2,clk=>clkd2,q=>ld2);
    dff_v(data=>d3,clk=>clkd3,q=>ld3);
    dff_v(data=>d4,clk=>clkd4,q=>ld4);

    dff_v(data=>c1,clk=>clkc1,q=>lc1);
    dff_v(data=>c2,clk=>clkc2,q=>lc2);
    dff_v(data=>c3,clk=>clkc3,q=>lc3);
    dff_v(data=>c4,clk=>clkc4,q=>lc4);

    dff(data=>acc,clk=>clk,q=>lacc);
    dff(data=>lacc,clk=>clk,q=>lacc2);

    dff(data=>fsel,clk=>clk,q=>lfsel);
    dff(data=>lfsel,clk=>clk,q=>lfsel2);
    dff(data=>lfsel2,clk=>clk,q=>lfsel3);
    dff(data=>lfsel3,clk=>clk,q=>lfsel4);


--  clock setup and hold checks
assert1: for i in 0 to 9 generate
x : ASSERT not((clk='1' and not clk'stable) and (not d1(i)'stable(tS))) 
        REPORT "setup violation on d1"
        SEVERITY warning;

y : ASSERT not((clk='1' and d1(i)'event) and (not clk'stable(tH))) 
        REPORT "hold violation on d1"
        SEVERITY warning;

end generate;

assert2: for i in 0 to 9 generate
x : ASSERT not((clk='1' and not clk'stable) and (not d2(i)'stable(tS))) 
        REPORT "setup violation on d2"
        SEVERITY warning;

y : ASSERT not((clk='1' and d2(i)'event) and (not clk'stable(tH))) 
        REPORT "hold violation on d2"
        SEVERITY warning;

end generate;

assert3: for i in 0 to 9 generate
x : ASSERT not((clk='1' and not clk'stable) and (not d3(i)'stable(tS))) 
        REPORT "setup violation on d3"
        SEVERITY warning;

y : ASSERT not((clk='1' and d3(i)'event) and (not clk'stable(tH))) 
        REPORT "hold violation on d3"
        SEVERITY warning;

end generate;

assert4: for i in 0 to 9 generate
x : ASSERT not((clk='1' and not clk'stable) and (not d4(i)'stable(tS))) 
        REPORT "setup violation on d4"
        SEVERITY warning;

y : ASSERT not((clk='1' and d4(i)'event) and (not clk'stable(tH))) 
        REPORT "hold violation on d4"
        SEVERITY warning;

end generate;

assert5: for i in 0 to 10 generate
x : ASSERT not((clk='1' and not clk'stable) and (not c1(i)'stable(tS))) 
        REPORT "setup violation on c1"
        SEVERITY warning;

y : ASSERT not((clk='1' and c1(i)'event) and (not clk'stable(tH))) 
        REPORT "hold violation on c1"
        SEVERITY warning;

end generate;

assert6: for i in 0 to 10 generate
x : ASSERT not((clk='1' and not clk'stable) and (not c2(i)'stable(tS))) 
        REPORT "setup violation on c2"
        SEVERITY warning;

y : ASSERT not((clk='1' and c2(i)'event) and (not clk'stable(tH))) 
        REPORT "hold violation on c2"
        SEVERITY warning;

end generate;

assert7: for i in 0 to 10 generate
x : ASSERT not((clk='1' and not clk'stable) and (not c3(i)'stable(tS))) 
        REPORT "setup violation on c3"
        SEVERITY warning;

y : ASSERT not((clk='1' and c3(i)'event) and (not clk'stable(tH))) 
        REPORT "hold violation on c3"
        SEVERITY warning;

end generate;

assert8: for i in 0 to 10 generate
x : ASSERT not((clk='1' and not clk'stable) and (not c4(i)'stable(tS))) 
        REPORT "setup violation on c4"
        SEVERITY warning;

y : ASSERT not((clk='1' and c4(i)'event) and (not clk'stable(tH))) 
        REPORT "hold violation on c4"
        SEVERITY warning;

end generate;


a : ASSERT not((clk='1' and not clk'stable) and (not acc'stable(tS))) 
        REPORT "setup violation on acc"
        SEVERITY warning;

b : ASSERT not((clk='1' and acc'event) and (not clk'stable(tH))) 
        REPORT "hold violation on acc"
        SEVERITY warning;

c : ASSERT not((clk='1' and not clk'stable) and (not fsel'stable(tS))) 
        REPORT "setup violation on fsel"
        SEVERITY warning;

d : ASSERT not((clk='1' and fsel'event) and (not clk'stable(tH))) 
        REPORT "hold violation on fsel"
        SEVERITY warning;

e : ASSERT not((clk='1' and not clk'stable) and (not ensel'stable(tS))) 
        REPORT "setup violation on ensel"
        SEVERITY warning;

f : ASSERT not((clk='1' and ensel'event) and (not clk'stable(tH))) 
        REPORT "hold violation on ensel"
        SEVERITY warning;

g1 : ASSERT not((clk='1' and not clk'stable) and (not enb1'stable(tS))) 
        REPORT "setup violation on enb1"
        SEVERITY warning;

h1 : ASSERT not((clk='1' and enb1'event) and (not clk'stable(tH))) 
        REPORT "hold violation on enb1"
        SEVERITY warning;

g2 : ASSERT not((clk='1' and not clk'stable) and (not enb2'stable(tS))) 
        REPORT "setup violation on enb2"
        SEVERITY warning;

h2 : ASSERT not((clk='1' and enb2'event) and (not clk'stable(tH))) 
        REPORT "hold violation on enb2"
        SEVERITY warning;

g3 : ASSERT not((clk='1' and not clk'stable) and (not enb3'stable(tS))) 
        REPORT "setup violation on enb3"
        SEVERITY warning;

h3 : ASSERT not((clk='1' and enb3'event) and (not clk'stable(tH))) 
        REPORT "hold violation on enb3"
        SEVERITY warning;

g4 : ASSERT not((clk='1' and not clk'stable) and (not enb4'stable(tS))) 
        REPORT "setup violation on enb4"
        SEVERITY warning;

h4 : ASSERT not((clk='1' and enb4'event) and (not clk'stable(tH))) 
        REPORT "hold violation on enb4"
        SEVERITY warning;


proc_m: process
    -- function evec2int (l: std_ulogic_vector)                      return natural;
    -- Translate signed vector into integer representation
    ---
       function evec2sint (l: std_logic_vector)                      
             return integer is
          variable result: integer := 0;
           attribute synthesis_return of result:variable is "FEED_THROUGH" ;
       begin
		  if ((l(l'left)='1') or (l(l'left)='H')) then -- negative
            for t1 in l'left-1 downto l'right loop
               result := result * 2;
               if (l(t1) = '0' or l(t1) = 'L') then
                  result := result - 1;
               end if;
            end loop;
			result := result-1;
		  else -- positive
            for t1 in l'left-1 downto l'right loop
               result := result * 2;
               if (l(t1) = '1' or l(t1) = 'H') then
                  result := result + 1;
               end if;
            end loop;
		  end if;
          return result;
       end evec2sint;
begin
    wait until rising_edge(clk);
	temp <=	evec2sint(ld1);
    ldtc1 <= evec2sint(ld1)*evec2sint(lc1);
    ldtc2 <= evec2sint(ld2)*evec2sint(lc2);
    ldtc3 <= evec2sint(ld3)*evec2sint(lc3);
    ldtc4 <= evec2sint(ld4)*evec2sint(lc4);
    rld1 <= real(evec2sint(ld1));
    rld2 <= real(evec2sint(ld2));
    rld3 <= real(evec2sint(ld3));
    rld4 <= real(evec2sint(ld4));
    rlc1 <= real(evec2sint(lc1));
    rlc2 <= real(evec2sint(lc2));
    rlc3 <= real(evec2sint(lc3));
    rlc4 <= real(evec2sint(lc4));
    rldtc1 <= (rld1/512.0)*(rlc1/512.0);
    rldtc2 <= (rld2/512.0)*(rlc2/512.0);
    rldtc3 <= (rld3/512.0)*(rlc3/512.0);
    rldtc4 <= (rld4/512.0)*(rlc4/512.0);
    

end process proc_m;

proc_acc: process
begin
    wait until rising_edge(clk);
    if (lacc2='1') then
      isum <= ldtc1+ldtc2+ldtc3+ldtc4+isum;
      risum <= rldtc1+rldtc2+rldtc3+rldtc4+risum;
	else -- no accumulate
      isum <= ldtc1+ldtc2+ldtc3+ldtc4;
      risum <= rldtc1+rldtc2+rldtc3+rldtc4;
    end if;
end process proc_acc;

proc_acc2: process
begin
    wait until rising_edge(clk);
    if (not_ocen='0') then
      isum2 <= isum;
      risum2 <= risum;
    end if;
end process proc_acc2;

proc_acc3: process
begin
    wait until rising_edge(clk);
    if (lfsel4='1') then
      isum3 <= isum2;
    else -- fractional mode
      isum3 <= integer((risum2*512.0));
    end if;
end process proc_acc3;

proc_out: process (not_oen,isum3)
begin
    if (not_oen='0') then
      s <= int2evec(isum3)(15 downto 0) after tD;
    else
      for i in 0 to 15 loop
        s(i) <= 'Z' after tDIS;
      end loop;
    end if;
end process proc_out;
  
end operate_fir;
